@@ -41,6 +41,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <sys/select.h>
 #include <utime.h>
 
@@ -59,32 +60,41 @@
 #define _CURL_MINIMUM_XFER_TIMEOUT	((long)(60))		/* 1 minute */
 #define _CURL_MINIMUM_PROGRESS_INTERVAL ((double)(0.2)) /* 0.2 seconds */
 
+#if defined CURLOPT_ACCEPT_ENCODING
+#define _CURL_ENCODING CURLOPT_ACCEPT_ENCODING
+#else
+#define _CURL_ENCODING CURLOPT_ENCODING
+#endif
+
 /* ========================================================================= **
  * Definitions
  * ========================================================================= */
 
 /* ------------------------------------------------------------------------- **
- * Global cURL handles
+ * Global curl handles
  * ------------------------------------------------------------------------- */
-/* If we want to use TclX' signal handling mechanism we need cURL to return
+/* If we want to use TclX' signal handling mechanism we need curl to return
  * control to our code from time to time so we can call Tcl_AsyncInvoke to
  * process pending signals. To do that, we could either abuse the curl progress
  * callback (which would mean we could no longer use the default curl progress
- * callback, or we need to use the cURL multi API. */
+ * callback, or we need to use the curl multi API. */
 static CURLM* theMHandle = NULL;
 /* We use a single global handle rather than creating and destroying handles to
  * take advantage of HTTP pipelining, especially to the packages servers. */
 static CURL* theHandle = NULL;
 
+/* Global libcurl version info. */
+static curl_version_info_data *libcurl_version_info = NULL;
+
 /* ------------------------------------------------------------------------- **
  * Prototypes
  * ------------------------------------------------------------------------- */
 int SetResultFromCurlErrorCode(Tcl_Interp* interp, CURLcode inErrorCode);
 int SetResultFromCurlMErrorCode(Tcl_Interp* interp, CURLMcode inErrorCode);
-int CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
+int CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
 
 typedef struct {
 	Tcl_Interp *interp;
@@ -96,6 +106,7 @@ static int CurlProgressHandler(tcl_callback_t *callback, double dltotal, double
 static void CurlProgressCleanup(tcl_callback_t *callback);
 
 void CurlInit(void);
+static void set_curl_version_info(void);
 
 /* ========================================================================= **
  * Entry points
@@ -147,23 +158,84 @@ SetResultFromCurlMErrorCode(Tcl_Interp *interp, CURLMcode inErrorCode)
 	return result;
 }
 
+/* Bug affecting some versions of Monterey through Sequoia:
+   reusing an easy handle for different protocols can cause a
+   crash. https://github.com/curl/curl/issues/13731
+   TODO: Add upper bound here when we know what version macOS 16 ships. */
+#if LIBCURL_VERSION_NUM >= 0x074d00
+static void cleanup_handle_if_needed(const char *theURL);
+
+static void cleanup_handle_if_needed(const char *theURL)
+{
+    static CURLU *currentURLHandle = NULL;
+    static CURLU *previousURLHandle = NULL;
+
+    set_curl_version_info();
+    if (libcurl_version_info == NULL) {
+        fprintf(stderr, "Warning: set_curl_version_info failed\n");
+        return;
+    }
+
+    if (libcurl_version_info->version_num >= 0x080600 && libcurl_version_info->version_num < 0x080800) {
+        if (previousURLHandle != NULL) {
+            curl_url_cleanup(previousURLHandle);
+        }
+        previousURLHandle = currentURLHandle;
+        currentURLHandle = curl_url();
+         if (currentURLHandle != NULL
+            && curl_url_set(currentURLHandle, CURLUPART_URL, theURL, 0) != CURLUE_OK) {
+            curl_url_cleanup(currentURLHandle);
+            currentURLHandle = NULL;
+        }
+        if (theMHandle != NULL && currentURLHandle != NULL && previousURLHandle != NULL) {
+            char *current_scheme = NULL;
+            char *previous_scheme = NULL;
+            char *current_host = NULL;
+            char *previous_host = NULL;
+            int must_recreate_handle = 0;
+            if (curl_url_get(previousURLHandle, CURLUPART_SCHEME, &previous_scheme, 0) == CURLUE_OK
+                && curl_url_get(currentURLHandle, CURLUPART_SCHEME, &current_scheme, 0) == CURLUE_OK
+                && 0 != strcasecmp(previous_scheme, current_scheme)) {
+                must_recreate_handle = 1;
+            }
+            if (!must_recreate_handle
+                && curl_url_get(previousURLHandle, CURLUPART_HOST, &previous_host, 0) == CURLUE_OK
+                && curl_url_get(currentURLHandle, CURLUPART_HOST, &current_host, 0) == CURLUE_OK
+                && 0 != strcasecmp(previous_host, current_host)) {
+                must_recreate_handle = 1;
+            }
+            if (must_recreate_handle) {
+                /* deallocate the handle and start again */
+                curl_multi_cleanup(theMHandle);
+                theMHandle = NULL;
+            }
+        }
+    }
+}
+#endif /* LIBCURL_VERSION_NUM >= 0x074d00 */
+
 /**
  * curl fetch subcommand entry point.
  *
- * syntax: curl fetch [--disable-epsv] [--ignore-ssl-cert] [--remote-time] [-u userpass] [--effective-url lasturlvar] [--progress "builtin"|callback] url filename
+ * syntax: curl fetch [--disable-epsv] [--ignore-ssl-cert] [--remote-time] [-u userpass] [--effective-url lasturlvar] [--progress "builtin"|callback] [--enable-compression] url filename
  *
  * @param interp		current interpreter
  * @param objc			number of parameters
  * @param objv			parameters
  */
 int
-CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
 	bool handleAdded = false;
 	FILE* theFile = NULL;
 	char theErrorString[CURL_ERROR_SIZE];
 
+	/* Always 0-initialize the error string, since older curl versions may not
+	 * initialize the error string buffer at all. See
+	 * https://trac.macports.org/ticket/60581. */
+	theErrorString[0] = '\0';
+
 	do {
 		int noprogress = 1;
 		int useepsv = 1;
@@ -178,6 +250,9 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		};
 		char* effectiveURL = NULL;
 		char* userAgent = PACKAGE_NAME "/" PACKAGE_VERSION " libcurl/" LIBCURL_VERSION;
+		const int MAXHTTPHEADERS = 100;
+		int numHTTPHeaders = 0;
+		const char* httpHeaders[MAXHTTPHEADERS];
 		int optioncrsr;
 		int lastoption;
 		const char* theURL;
@@ -188,6 +263,7 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		struct curl_slist *headers = NULL;
 		struct CURLMsg *info = NULL;
 		int running; /* number of running transfers */
+		char* acceptEncoding = NULL;
 
 		/* we might have options and then the url and the file */
 		/* let's process the options first */
@@ -240,6 +316,26 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 					theResult = TCL_ERROR;
 					break;
 				}
+			} else if (strcmp(theOption, "--append-http-header") == 0) {
+				/* check we also have the parameter */
+				if (optioncrsr < lastoption) {
+					optioncrsr++;
+					if (numHTTPHeaders < MAXHTTPHEADERS) {
+						httpHeaders[numHTTPHeaders++] = Tcl_GetString(objv[optioncrsr]);
+					} else {
+						Tcl_SetResult(interp,
+							"curl fetch: Too many --append-http-header options",
+							TCL_STATIC);
+						theResult = TCL_ERROR;
+						break;
+					}
+				} else {
+					Tcl_SetResult(interp,
+						"curl fetch: --append-http-header option requires a parameter",
+						TCL_STATIC);
+					theResult = TCL_ERROR;
+					break;
+				}
 			} else if (strcmp(theOption, "--progress") == 0) {
 				/* check we also have the parameter */
 				if (optioncrsr < lastoption) {
@@ -253,6 +349,8 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 					theResult = TCL_ERROR;
 					break;
 				}
+			} else if (strcmp(theOption, "--enable-compression") == 0) {
+				acceptEncoding = "";
 			} else {
 				Tcl_ResetResult(interp);
 				Tcl_AppendResult(interp, "curl fetch: unknown option ", theOption, NULL);
@@ -291,6 +389,10 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			break;
 		}
 
+#if LIBCURL_VERSION_NUM >= 0x074d00
+		cleanup_handle_if_needed(theURL);
+#endif
+
 		/* Create the CURL handles */
 		if (theMHandle == NULL) {
 			/* Re-use existing multi handle if theMHandle isn't NULL */
@@ -464,8 +566,25 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			}
 		}
 
+		/* a CURLOPT_ACCEPT_ENCODING of "" means to let cURL write the
+		 * Accept-Encoding header for you, based on what the library
+		 * was compiled to support.
+		 * A value of NULL disables all attempts at decompressing responses.
+		*/
+#ifdef _CURL_ENCODING
+		theCurlCode = curl_easy_setopt(theHandle, _CURL_ENCODING, acceptEncoding);
+		if (theCurlCode != CURLE_OK) {
+			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+			break;
+		}
+#endif
+
 		/* Clear the Pragma: no-cache header */
 		headers = curl_slist_append(headers, "Pragma:");
+		/* Append any optional headers */
+		for (int iH = 0; iH < numHTTPHeaders; ++iH) {
+			headers = curl_slist_append(headers, httpHeaders[iH]);
+		}
 		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_HTTPHEADER, headers);
 		if (theCurlCode != CURLE_OK) {
 			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
@@ -594,10 +713,23 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			theResult = TCL_ERROR;
 			break;
 		}
-		
+
 		if (info->data.result != CURLE_OK) {
-			/* execution failed, use the error string */
-			Tcl_SetResult(interp, theErrorString, TCL_VOLATILE);
+			/* execution failed, use the error string if it is set */
+			if (theErrorString[0] != '\0') {
+				Tcl_SetResult(interp, theErrorString, TCL_VOLATILE);
+			} else {
+				/* When the error buffer does not hold useful information,
+				 * generate our own message. Use a larger buffer since we add
+				 * a significant amount of text. */
+				char errbuf[256 + CURL_ERROR_SIZE];
+				snprintf(errbuf, sizeof(errbuf),
+					"curl_multi_info_read() returned {.msg = CURLMSG_DONE, "
+					".data.result = %d (!= CURLE_OK)}, but the error buffer "
+					"is not set. curl_easy_strerror(.data.result): %s",
+					info->data.result, curl_easy_strerror(info->data.result));
+				Tcl_SetResult(interp, errbuf, TCL_VOLATILE);
+			}
 			theResult = TCL_ERROR;
 			break;
 		}
@@ -606,7 +738,7 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		(void) fclose(theFile);
 		theFile = NULL;
 
-#if LIBCURL_VERSION_NUM == 0x070d01 /* work around broken Tiger version of cURL */
+#if LIBCURL_VERSION_NUM == 0x070d01 /* work around broken Tiger version of curl */
 		if (remotetime) {
 			FILE *fp;
 			char *tmp, *p;
@@ -689,7 +821,7 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
 	FILE* theFile = NULL;
@@ -703,6 +835,7 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		CURLcode theCurlCode;
 		long theModDate;
 		long userModDate;
+		const char* userAgent = PACKAGE_NAME "/" PACKAGE_VERSION " libcurl/" LIBCURL_VERSION;
 
 		optioncrsr = 2;
 		lastoption = objc - 3;
@@ -773,6 +906,13 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			break;
 		}
 
+		/* -A option */
+		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_USERAGENT, userAgent);
+		if (theCurlCode != CURLE_OK) {
+			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+			break;
+		}
+
 		/* --max-redirs option, same default as curl command line */
 		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_MAXREDIRS, 50);
 		if (theCurlCode != CURLE_OK) {
@@ -925,7 +1065,7 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
 	FILE* theFile = NULL;
@@ -1133,7 +1273,7 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
 	FILE* theFile = NULL;
@@ -1368,6 +1508,14 @@ CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 	return theResult;
 }
 
+/* Fill in global libcurl_version_info structure */
+static void set_curl_version_info(void)
+{
+    if (libcurl_version_info == NULL) {
+        libcurl_version_info = curl_version_info(CURLVERSION_NOW);
+    }
+}
+
 /**
  * curl command entry point.
  *
@@ -1381,7 +1529,7 @@ CurlCmd(
 		ClientData clientData UNUSED,
 		Tcl_Interp* interp,
 		int objc,
-		Tcl_Obj* CONST objv[])
+		Tcl_Obj* const objv[])
 {
 	typedef enum {
 		kCurlFetch,
@@ -1437,7 +1585,7 @@ CurlCmd(
  * libcurl will never be cleaned (where should I plug the hook?)
  */
 void
-CurlInit()
+CurlInit(void)
 {
 	curl_global_init(CURL_GLOBAL_ALL);
 }
